/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/
#ifndef __INTERPRETER_H__
#define __INTERPRETER_H__

#include "boolean.h"
#include "memory.h"
#include "types.h"
#include "vm.h"

/** Defines the supported assembler instructions.
 */
typedef enum {

  /** Assembler instruction:
   *  NOP
   *
   *  Do nothing but set the corresponding 
   *  memory address to 0.
   *  See the instruction NOOP for an instruction
   *  which does not change the memory state.
   */
  ASM_NOP=0,

  /** Assembler instruction:
   *  NOOP
   *
   *  Do nothing but does not set the corresponding 
   *  memory address to 0.
   *  See the instruction NOP for an instruction
   *  which changes the memory state.
   */
  ASM_NOOP,

  /** Assembler instruction:
   *  DECLARE n
   *
   *  Declare a set of register.
   *  A register could be addressed
   *  with "Rx" where "x" is the number
   *  of the register. Any previous call
   *  to DECLARE are automatically destroyed.
   *
   * @see #F_DECLARE()
   */
  ASM_DECLARE,

  /** Assembler instruction:
   *  MOVE address value
   *
   *  Set the content of the specified address
   *  with the given value.
   *
   * @see #F_MOVE()
   */
  ASM_MOVE,

  /** Assembler instruction:
   *  MOVA address address
   *
   *  Set the content of the first parameter
   *  with the content of second parameter.
   *
   * @see #F_MOVA()
   */
  ASM_MOVA,

  /** Assembler instruction:
   *  LESSER address address
   *
   *  Test if the content of the first address is
   *  lesser than the content of the second.
   *
   * @see #F_LESSER()
   */
  ASM_LESSER,

  /** Assembler instruction:
   *  GREATER address address
   *
   *  Test if the content of the first address is
   *  greater than the content of the second.
   *
   * @see #F_GREATER()
   */
  ASM_GREATER,

  /** Assembler instruction:
   *  EQUAL address address
   *
   *  Test if the content of the first address is
   *  equal to the content of the second.
   *
   * @see #F_EQUAL()
   */
  ASM_EQUAL,

  /** Assembler instruction:
   *  ZERO address
   *
   *  Test if the content of the address is
   *  equal to zero.
   *
   * @see #F_ZERO()
   */
  ASM_ZERO,

  /** Assembler instruction:
   *  JUMP address
   *
   *  Goto the specified address
   *
   * @see #F_JUMP()
   */
  ASM_JUMP,

  /** Assembler instruction:
   *  JUMPIF address
   *
   *  Goto the specified address if the
   *  current test flag was true.
   *
   * @see #F_JUMPIF()
   */
  ASM_JUMPIF,

  /** Assembler instruction:
   *  JUMPIFNOT address
   *
   *  Goto the specified address if the
   *  current test flag was false.
   *
   * @see #F_JUMPIFNOT()
   */
  ASM_JUMPIFNOT,

  /** Assembler instruction:
   *  INC address
   *
   *  Incrementes by 1.
   *
   * @see #F_INC()
   */
  ASM_INC,

  /** Assembler instruction:
   *  DEC address
   *
   *  Decrementes by 1.
   *
   * @see #F_DEC()
   */
  ASM_DEC,

  /** Assembler instruction:
   *  ADD address address
   *
   *  Adds the content of the second address
   *  to the content of the first.
   *
   * @see #F_ADD()
   */
  ASM_ADD,

  /** Assembler instruction:
   *  SUB address address
   *
   *  Substracts the content of the second address
   *  to the content of the first.
   *
   * @see #F_SUB()
   */
  ASM_SUB,

  /** Assembler instruction:
   *  MUL address address
   *
   *  Multiply the content of the second address
   *  with the content of the first and put the
   *  result at the first address.
   *
   * @see #F_MUL()
   */
  ASM_MUL,

  /** Assembler instruction:
   *  DIV address address
   *
   *  Divide the content of the first address
   *  with the content of the second and put the
   *  result at the first address.
   *
   * @see #F_DIV()
   */
  ASM_DIV,

  /** Assembler instruction:
   *  MOD address address
   *
   *  Divide the content of the first address
   *  with the content of the second and put the
   *  modulo at the first address.
   *
   * @see #F_MOD()
   */
  ASM_MOD,

  /** Assembler instruction:
   *  RAND address
   *
   *  Generates a random number between 0 and the
   *  value atthe specified address and put the result
   *  at the specified address.
   *
   * @see #F_RAND()
   */
  ASM_RAND,

  /** Assembler instruction:
   *  MEMSIZE address
   *
   *  Put the size of the memory into the
   *  specified address. If the size neads
   *  more than once memory element, the
   *  following are set (but this operation
   *  never generates any memory access error
   *  if it was not enough space to put the size).
   *
   * @see #F_MEMSIZE()
   */
  ASM_MEMSIZE,

  /** Assembler instruction:
   *  CODESEG address address
   *
   *  Put the starting and the ending addresses
   *  of the code segment.
   *
   * @see #F_CODESEG()
   */
  ASM_CODESEG,

  /** Assembler instruction:
   *  REGSEG address address
   *
   *  Put the starting and the ending addresses
   *  of the registry segment.
   *
   * @see #F_REGSEG()
   */
  ASM_REGSEG,

  /** Assembler instruction:
   *  CO address
   *
   *  Put the value of the ordinal counter of the current
   *  opponent at the specified address.
   *
   * @see #F_CO()
   */
  ASM_CO,

  /** Assembler instruction:
   *  DISPVAL value
   *
   *  Display the specified value.
   *
   * @see #F_DISPVAL()
   */
  ASM_DISPVAL,

  /** Assembler instruction:
   *  DISP address
   *
   *  Display the specified value at the specified address.
   *
   * @see #F_DISP()
   */
  ASM_DISP,

  /** Assembler instruction:
   *  DISPSTR address count
   *
   *  Display the set of values as characters
   *
   * @see #F_DISPSTR()
   */
  ASM_DISPSTR,

  /** Assembler instruction:
   *  END
   *
   *  Stop the program.
   *
   * @see #F_END()
   */
  ASM_END,

  /** Quantity of assembler instructions.
   */
  ASSEMBLER_SIZE,

} ASSEMBLER_CODE ;

/** Type for a assembler instruction's parameter.
 */
typedef MEMORY_ITEM ASSEMBLER_PARAM ;

/** Size of an assembler instruction (in memory elements)
 */
#define ASM_INSTRUCTION_SIZE 3

/** No parameter.
 */
#define PASM_NONE 0

/** One parameter.
 */
#define PASM_ONE  1

/** Two parameters.
 */
#define PASM_TWO  2

/** First parameter is an address.
 */
#define PASM_ADR1 4

/** Second parameter is an address.
 */
#define PASM_ADR2 8

/** Run the specified assembler instruction.
 *
 * @param vm is the virtual machine inside which the instruction
 *        must be ran
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_run(VIRTUAL_MACHINE* vm, ASSEMBLER_CODE inst, ASSEMBLER_PARAM param1, ASSEMBLER_PARAM param2) ;

/** Translates the addresses of the specified instruction.
 * It is needed because the addreses of a program are
 * always given from the start of the program code.
 * This translation permits to update the addresses to
 * be compliant with the real location of the source code.
 *
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @param base is the base address which corresponds to 0
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_translateAddresses(ASSEMBLER_CODE inst, ASSEMBLER_PARAM* param1, ASSEMBLER_PARAM* param2, MEMORY_ADDRESS base) ;

/** Replies the identifier that corresponds to the
 *  assembler instruction which is passed in parameter.
 *
 * @param token is the string which corresponds to the instruction to search
 * @param inst is the identifier of the instruction
 * @param proto is the prototype of the found instruction
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_getInstructionDef(char* token, ASSEMBLER_CODE* inst, unsigned char* proto) ;

/** Replies the prototype of the specified assembler function.
 *
 * @param inst is the identifier of the instruction
 * @return the prototype of 0
 */
unsigned char ASM_getInstructionPrototype(ASSEMBLER_CODE inst) ;

/** Adds an opponents to the specified opponent's list.
 *
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param count is the quantity of opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponent(MEMORY_ITEM*** codes, CODE_SIZE** sizes, OPPONENT_SIZE* count) ;

/** Adds an opponent's instruction into the specified opponent's code.
 *
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param count is the quantity of opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponentCode(ASSEMBLER_CODE inst, ASSEMBLER_PARAM param1, ASSEMBLER_PARAM param2, MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE count) ;

/** Adds an opponent's data into the specified opponent's code.
 *
 * @param data is the data to add
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param nopponent is the opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponentData(ASSEMBLER_PARAM data, MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE nopponent) ;

/** Display the codes of the opponents.
 *
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param count is the quantity of opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
void ASM_displayCodes(MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE count) ;

#endif
