/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#ifndef __INTERPRETER_C__
#error Can not compile this file outside interpreter.c
#endif

/** Declare a set of register.
 *  A register could be addressed
 *  with "Rx" where "x" is the number
 *  of the register. Any previous call
 *  to DECLARE are automatically destroyed.
 *
 * @param vm is the current virtual machine.
 * @param n is the quantity of register to declare
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(DECLARE, n) {
  if (n<=0) {
    ASMERROR(ERR_ASM_NULLPARAMETER) ;
  }
  else if ( ! VM_computeArrayBaseAddress(vm,n) ) {
    ASMERROR(ERR_ASM_NOTENOUGHMEMORY) ;
  }
  else {
    return TRUE ;
  }  
  return FALSE ;
} PASM_END

/** Goto the specified address.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(JUMP, adr) {
  VM_setOrdinalCounter(vm,adr-ASM_INSTRUCTION_SIZE) ;
  return TRUE ;
} PASM_END

/** Goto the specified address if the current
 *  test flag was true.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(JUMPIF, adr) {
  boolean b ;
  if (!VM_getLastTest(vm,&b)) {
    ASMERROR(ERR_ASM_NOTPREVIOUSTEST) ;
    return FALSE ;
  }
  if (b) {
    VM_setOrdinalCounter(vm,adr-ASM_INSTRUCTION_SIZE) ;
  }
  return TRUE ;
} PASM_END

/** Goto the specified address if the current
 *  test flag was false.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(JUMPIFNOT, adr) {
  boolean b ;
  if (!VM_getLastTest(vm,&b)) {
    ASMERROR(ERR_ASM_NOTPREVIOUSTEST) ;
    return FALSE ;
  }
  if (!b) {
    VM_setOrdinalCounter(vm,adr-ASM_INSTRUCTION_SIZE) ;
  }
  return TRUE ;
} PASM_END

/** Set the content of the specified address
 *  with the given value.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to set.
 * @param value is the value to copy.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(MOVE, adr, value) {
  return MEM_set(mem,adr,value) ;
} PASM_END

/** Set the content of the first parameter
 *  with the content of second parameter.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to set.
 * @param adr2 is the address to copy.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(MOVA, adr1, adr2) {
  MEMORY_ITEM value = MEM_get(mem,adr2) ;
  return MEM_set(mem,adr1,value) ;
} PASM_END

/** Display the specified value.
 *
 * @param vm is the current virtual machine.
 * @param value is the value to display.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(DISPVAL, value) {
  display(stdout, "*** opponent #%lu: %lu\n",
	  (unsigned long)vm->current_player,
	  (unsigned long)value) ;
  return TRUE ;
} PASM_END

/** Display the value at the specified address.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to display.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(DISP, adr) {
  MEMORY_ITEM value = MEM_get(mem,adr) ;
  display(stdout, "*** opponent #%lu: %lu\n",
	  (unsigned long)vm->current_player,
	  (unsigned long)value) ;
  return TRUE ;
} PASM_END

/** Display a set of character.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address where the string start.
 * @param count is the quantity of character to display.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(DISPSTR, adr, count) {
  unsigned long i ;
  display(stdout, "*** opponent #%lu: ",
	  (unsigned long)vm->current_player) ;
  for(i=0; i<count; i++) {
    MEMORY_ITEM value ;
    if (!MEM_isValidAdr(mem,adr+i)) {
      ASMERROR(ERR_ASM_INVALIDADDRESS) ;
      return FALSE ;
    }
    value = MEM_get(mem,adr+i) ;
    display(stdout, "%c", (char)value) ;
  }
  display(stdout, "\n");
  return TRUE ;
} PASM_END

/** Stop this program.
 *
 * @param vm is the current virtual machine.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF(END) {
  boolean b = VM_stopOpponent(vm) ;
  if (b) {
    ASMERRORNOLOG(ERR_ASM_END) ;
  }
  return b ;
} PASM_END

/** Test if the content of the first address is
 *  lesser than the content of the second.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the first address.
 * @param adr2 is the second address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(LESSER, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return VM_setLastTest(vm,(value1<value2)) ;
} PASM_END

/** Test if the content of the first address is
 *  greater than the content of the second.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the first address.
 * @param adr2 is the second address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(GREATER, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return VM_setLastTest(vm,(value1>value2)) ;
} PASM_END

/** Test if the content of the first address is
 *  equal to the content of the second.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the first address.
 * @param adr2 is the second address.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(EQUAL, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return VM_setLastTest(vm,(value1==value2)) ;
} PASM_END

/** Test if the content of the first address is
 *  equal to zero.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to test.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(ZERO, adr) {
  MEMORY_ITEM value = MEM_get(mem,adr) ;
  return VM_setLastTest(vm,(value==0)) ;
} PASM_END

/** Incrementes by 1.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(INC, adr) {
  MEMORY_ITEM value = MEM_get(mem,adr) ;
  return MEM_set(mem,adr,value+1) ;
} PASM_END

/** Decrementes by 1.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(DEC, adr) {
  MEMORY_ITEM value = MEM_get(mem,adr) ;
  return MEM_set(mem,adr,value-1) ;
} PASM_END

/** Adds the content of the second address
 *  to the content of the first.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to update.
 * @param adr2 is the address of the second operand.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(ADD, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return MEM_set(mem,adr1,value1+value2) ;  
} PASM_END

/** Substracts the content of the second address
 *  to the content of the first.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to update.
 * @param adr2 is the address of the second operand.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(SUB, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return MEM_set(mem,adr1,value1-value2) ;  
} PASM_END

/** Multiply the content of the second address
 *  with the content of the first and put the
 *  result at the first address.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to update.
 * @param adr2 is the address of the second operand.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(MUL, adr1, adr2) {
  MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  return MEM_set(mem,adr1,value1*value2) ;  
} PASM_END

/** Divide the content of the first address
 *  with the content of the second and put the
 *  result at the first address.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to update.
 * @param adr2 is the address of the second operand.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(DIV, adr1, adr2) {
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  if ( value2 != 0 ) {
    MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
    return MEM_set(mem,adr1,value1/value2) ;
  }
  else {
    ASMERROR(ERR_ASM_DIVISIONBYZERO) ;
    return FALSE ;
  }
} PASM_END

/** Divide the content of the first address
 *  with the content of the second and put the
 *  modulo at the first address.
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the address to update.
 * @param adr2 is the address of the second operand.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(MOD, adr1, adr2) {
  MEMORY_ITEM value2 = MEM_get(mem,adr2) ;
  if ( value2 != 0 ) {
    MEMORY_ITEM value1 = MEM_get(mem,adr1) ;
    return MEM_set(mem,adr1,value1%value2) ;
  }
  else {
    ASMERROR(ERR_ASM_DIVISIONBYZERO) ;
    return FALSE ;
  }
} PASM_END

/** Generates a random number between 0 and the
 *  value atthe specified address and put the result
 *  at the specified address.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(RAND, adr) {
  MEMORY_ITEM max = MEM_get(mem,adr) ;
  int r = rand() ;
  return MEM_set(mem, adr, (MEMORY_ITEM)((float)r / (float)RAND_MAX * (float)max)) ;
} PASM_END

/** Put the size of the memory into the
 *  specified address. If the size neads
 *  more than once memory element, the
 *  following are set (but this operation
 *  never generates any memory access error
 *  if it was not enough space to put the size).
 *
 * @param vm is the current virtual machine.
 * @param adr is the address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(MEMSIZE, adr) {
  MEMORY_SIZE size = MEM_size(mem) ;
  unsigned long i ;
  MEMORY_ITEM flag = 0 ;
  float n = (float)sizeof(MEMORY_SIZE) / (float)sizeof(MEMORY_ITEM) ;
  if ( n > 1.0 ) {
    unsigned char j ;
    for(j=0;j<sizeof(MEMORY_ITEM); j++) {
      flag = (flag<<1)|1 ;
    }
  
    for(i=0; i<(unsigned long)n; i++) {
      MEMORY_ITEM it = flag & size ;
      if ( ! MEM_isValidAdr(mem,adr+(unsigned long)n-i) ) return TRUE ;
      if ( ! MEM_set(mem,adr+(unsigned long)n-i,it) ) return FALSE ;
      size >>= sizeof(MEMORY_ITEM) ;
    }
  }
  return MEM_set(mem, adr, (MEMORY_ITEM)size) ;
} PASM_END

/** Put the starting and the ending addresses
 *  of the code segment. 
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the first address to update.
 * @param adr2 is the second address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(CODESEG, adr1, adr2) {
  MEMORY_ADDRESS startadr, endadr;
  if ( ! VM_getCodeSegmentAddresses(vm,&startadr,&endadr) ) {
    ASMERROR(ERR_ASM_NOCODESEGMENT);
    return FALSE ;
  }
  return ( (MEM_set(mem,adr1,startadr)) &&
	   (MEM_set(mem,adr2,endadr)) ) ;
} PASM_END

/** Put the starting and the ending addresses
 *  of the registry segment. 
 *
 * @param vm is the current virtual machine.
 * @param adr1 is the first address to update.
 * @param adr2 is the second address to update.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF2(REGSEG, adr1, adr2) {
  MEMORY_ADDRESS startadr, endadr;
  if ( ! VM_getArrayAddresses(vm,&startadr,&endadr) ) {
    ASMERROR(ERR_ASM_NODECLARE);
    return FALSE ;
  }
  return ( (MEM_set(mem,adr1,startadr)) &&
	   (MEM_set(mem,adr2,endadr)) ) ;
} PASM_END

/** Put the value of the ordinal counter of the current
 *  opponent at the specified address.
 *
 * @param vm is the current virtual machine.
 * @param adr is the address were to put the ordinal counter.
 * @return TRUE is the instruction end successfully, otherwhise FALSE
 */
PASM_DEF1(CO, adr) {
  MEMORY_ADDRESS co = VM_getOrdinalCounter(vm) ;
  return MEM_set(mem,adr,co) ;
} PASM_END
