/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#include <stdlib.h>
#include <stdio.h>
#include "display.h"
#include "vm.h"
#include "loader.h"

/** Size of the memory.
 */
#define MEMSIZE 2048

/** Loads the opponents scripts.
 *
 * @param vm is the virtual machine
 * @param nopponents is the quantity of opponents
 * @param files is the array of input streams
 */
OPPONENT_SIZE loading(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponents, FILE** files) {
  OPPONENT_SIZE n ;
  d_log("Loading opponents source codes\n") ;
  n = loadOpponentsFromFiles(vm,files,nopponents) ;
  if (n==nopponents) d_log("done\n") ;
  return n ;
}

/** Run the game.
 *
 * @param vm is the virtual machine
 * @param nopponents is the quantity of opponents
 */
void run(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponents) {
  OPPONENT_SIZE i ;

  if ( nopponents > 0 ) {
    d_log("Starting the war with %u opponent(s)\n",nopponents) ;

    if ( VM_run(vm) ) {
      d_log("Winner is the opponent %lu after %lu turn(s)\n",
	    (unsigned long)VM_getWinner(vm),
	    (unsigned long)VM_getTurns(vm)+1) ;
    }
    else {
      d_log("No winner\n") ;
    }

    /* display error messages */
    for(i=0; i<nopponents; i++) {
      if ( VM_getOrdinalCounterFor(vm,i) != VNULL ) {
	d_log("\topponent %u: winner\n", i);
      }
      else {
	const char* message = VM_getErrorMsg(vm,i) ;
	if ( message != NULL ) {
	  d_log("\topponent %u: %s\n", i, message);
	}
	else {
	  d_log("\topponent %u: don't know why it failed\n", i);
	}
      }
    }
  }
}

/** Main program.
 */
int main(int argc, char** argv) {
  VIRTUAL_MACHINE vm ;
  unsigned long i ;
  OPPONENT_SIZE nopponents=0 ;
  FILE** files=NULL ;

  if (argc<=1) {
    d_error("usage: %s script1 script2...\n", argv[0]) ;
    return -1 ;
  }

  d_log("eWar engine by Stephane GALLAND\n") ;

  /* open the files */
  for(i=1; i<argc; i++) {
    FILE* fid = fopen(argv[i],"rt") ;
    if ( fid == NULL ) {
      unsigned long j;
      d_error("error opening the file %s\n", argv[i]) ;
      if (files!=NULL) {
	for(j=0; j<nopponents; j++) {
	  fclose(files[j]) ;
	}
	free(files) ;
      }
      return -1 ;
    }
    else {
      files = (FILE**)realloc(files,sizeof(FILE*)*(nopponents+1)) ;
      files[nopponents] = fid ;
      nopponents ++ ;
    }
  }

  /* initializing */
  d_log("Initializing virtual machine....\n") ;
  VM_init(&vm,nopponents,MEMSIZE);
  d_log("done\n") ;

  /* loading */
  nopponents = loading(&vm,nopponents,files) ;

  /* close opened files */
  if (files!=NULL) {
    for(i=0; i<nopponents; i++) {
      fclose(files[i]) ;
    }
    free(files) ;
    files = NULL ;
  }

  /* running */
  run(&vm,nopponents) ;

  /* deleting */
  VM_delete(&vm);

  return 0 ;
}
