/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#include <stdlib.h>
#include <stdarg.h>
#include "memory.h"
#ifdef __DEBUG__
#include "display.h"
#endif

/** Initializes the specified virtual memory.
 *  This function must always be called before
 *  any operation on the memory.
 *
 * @param memory is the memory to initialize
 * @param size is the desired size (in items) of the memory
 */
void MEM_init(VIRTUAL_MEMORY* memory, MEMORY_SIZE size) {
  if ( (memory!=NULL) && (size>0)) {
#ifdef __DEBUG__
    d_log("\tmemory elements: %lu\n", (unsigned long)size) ;
    d_log("\telement's size: %lu byte(s)\n", (unsigned long)sizeof(MEMORY_ITEM)) ;
    d_log("\tmemory: %lu byte(s)\n", (unsigned long)(sizeof(MEMORY_ITEM)*size)) ;
#endif
    memory->content = (MEMORY_ITEM*)malloc(sizeof(MEMORY_ITEM)*size) ;
    memory->size = size ;
  }
}

/** Destroyes the specified virtual memory.
 *  This function must be called after all the
 *  operations on the specified memory.
 *
 * @param memory is the memory to delete
 */
void MEM_delete(VIRTUAL_MEMORY* memory) {
  if (memory!=NULL) {
    if ( memory->content != NULL ) {
      free(memory->content) ;
    }
    memory->content = NULL ;
    memory->size = 0 ;
  }
}

/** Replies if the specified memory address is
 *  valid for the specified memory.
 *  An address is valid only if it's corresponding
 *  to an item (data) of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address to check
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidAdr(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr) {
  return ( (memory!=NULL) &&
	   (memory->size>0) &&
	   (adr<memory->size) ) ;
}

/** Replies if the specified memory address is
 *  valid for an instruction storage.
 *
 * @param memory is the memory to use
 * @param count is the quantity of data to set
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidDataAdr(VIRTUAL_MEMORY* memory, MEMORY_SIZE count, ...) {
  va_list va;
  MEMORY_SIZE dec ;
  va_start(va,count);

  for(dec=0; dec<count; dec++) {
    if (!MEM_isValidAdr(memory,va_arg(va,MEMORY_ADDRESS))) {
      va_end(va);
      return FALSE ;
    }
  }

  va_end(va);
  return TRUE ;
}

/** Replies if the specified set of contigue memory addresses is
 *  valid.
 *
 * @param memory is the memory to use
 * @param adr is the address to check
 * @param count is the quantity of addresses
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidContigueAdr(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count) {
  MEMORY_SIZE dec ;

  for(dec=0; dec<count; dec++) {
    if (!MEM_isValidAdr(memory,adr+dec)) {
      return FALSE ;
    }
  }

  return TRUE ;
}

/** Replies if the size in bytes of the memory.
 *
 * @param memory is the memory to use
 * @return the size in bytes of the memory
 */
MEMORY_SIZE MEM_size(VIRTUAL_MEMORY* memory) {
  if (memory!=NULL) {
    return memory->size ;
  }
  return 0 ;
}

/** Put a data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param data is the data to put
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean MEM_set(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_ITEM value) {
  if ( MEM_isValidAdr(memory,adr) ) {
    memory->content[adr] = value ;
#ifdef __DEBUG__
    d_log("memset(%lu) = %lu\n",
	   (unsigned long)adr,
	   (unsigned long)value) ;
#endif
    return TRUE ;
  }
  return FALSE ;
}

/** Replies the data at the specified address of the memory.
 *
 * @param memory is the memory to use.
 * @param adr is the address at which the data must be put.
 * @return the data at the specified address or 0 if
 *         the specified address was wrong.
 */
MEMORY_ITEM MEM_get(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr) {
  if ( MEM_isValidAdr(memory,adr) ) {
#ifdef __DEBUG__
    d_log("memget(%lu) = %lu\n",
	  (unsigned long)adr,
	  (unsigned long)memory->content[adr]) ;
#endif
    return memory->content[adr] ;
  }
  return (MEMORY_ITEM)0 ;
}

/** Put a set of data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param count is the quantity of data to set
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean MEM_setData(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count, ...) {
  va_list va ;
  MEMORY_SIZE dec=0 ;

  if (!MEM_isValidContigueAdr(memory,adr,count)) return FALSE ;

  va_start(va, count);

#ifdef __DEBUG__
  d_log("memsetdata(%lu,%lu) = (",
	   (unsigned long)adr,
	   (unsigned long)count) ;
#endif

  for(dec=0;dec<count; dec++) {
    
    memory->content[adr+dec] = va_arg(va,MEMORY_ITEM) ;
#ifdef __DEBUG__
    d_log(((dec>0)?";%lu":"%lu"),(unsigned long)memory->content[adr+dec]) ;
#endif

  }

#ifdef __DEBUG__
  d_log(")\n") ;
#endif
  va_end(va);

  return TRUE ;
}

/** Replies the set of data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param count is the quantity of data to set
 * @return TRUE if something was replied, otherwhise FALSE
 */
boolean MEM_getData(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count, ...) {
  va_list va ;
  MEMORY_SIZE dec=0 ;

  if (!MEM_isValidContigueAdr(memory,adr,count)) return FALSE ;

  va_start(va, count);

#ifdef __DEBUG__
    d_log("memgetdata(%lu,%lu) = (",
	     (unsigned long)adr,
	     (unsigned long)count) ;
#endif

  for(dec=0;dec<count; dec++) {
    MEMORY_ITEM* var ;

    var = va_arg(va,MEMORY_ITEM*) ;

#ifdef __DEBUG__
    d_log(((dec>0)?";%lu":"%lu"),(unsigned long)memory->content[adr+dec]) ;
#endif

    if (var!=NULL) (*var) = memory->content[adr+dec] ;

  }

#ifdef __DEBUG__
  d_log(")\n") ;
#endif
  va_end(va);

  return TRUE ;
}

