/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/
#ifndef __EW_VM_H__
#define __EW_VM_H__

#include "boolean.h"
#include "memory.h"
#include "types.h"
#include "error.h"

/** Defines the maximal quantity of turns.
 */
#define VM_TURN_MAX 50000

/** Defines the data structure for a virtual machine.
 *  It includes:
 *  - virtual memory
 *  - quantity of opponents
 *  - ordinal counters for each opponent
 *  - last boolean tests' results for each opponent
 *  - last array declarations for each opponent
 *  - current player which is ran
 *  - list of error messages which explain why an
 *    opponent was stopped
 */
typedef struct {

  VIRTUAL_MEMORY memory ;

  unsigned char nopponents ;

  MEMORY_ADDRESS* ordinal_counters ;
  MEMORY_ADDRESS* array_decls ;
  MEMORY_ADDRESS* code_segments ;
  boolean* boolean_tests ;
  ERRORNUM* error_messages ;

  unsigned short current_player ;
  unsigned long long turns ;

} VIRTUAL_MACHINE ;

/** Initializes the virtual machine.
 *
 * @param vm is the virtual machine to initialize.
 * @param nopponents is the quantity of opponents supported by the virtual machine
 * @param memsize is the size in bytes of the virtual memory
 */
void VM_init(VIRTUAL_MACHINE* vm,OPPONENT_SIZE nopponents, MEMORY_SIZE memsize) ;

/** Destroyes the virtual machine.
 *
 * @param vm is the virtual machine to delete.
 */
void VM_delete(VIRTUAL_MACHINE* vm) ;

/** Loads the code of an opponent.
 *  This operation erase the previously loaded opponent
 *  at the same index.
 *  This method does not check if any other opponent has
 *  any source code where the loaded opponent was put.
 *
 * @param vm is the virtual machine to use.
 * @param nopponent is the index of the opponent to load.
 * @param code is the source code to load into the memory.
 * @param size is the source code's size.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_unsecureOpponentLoading(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent, MEMORY_ITEM* code, CODE_SIZE size) ;

/** Loads the code of all the opponents.
 *  This operation erase the previously loaded opponents.
 *  This method ensure all the proposed source code will not
 *  be put in same memory items.
 *
 * @param vm is the virtual machine to use.
 * @param nopponents is the quantity of opponents to load
 * @param codes is the array of source codes.
 * @param sizes is the array of source code's sizes.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_secureOpponentLoading(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent, MEMORY_ITEM** codes, CODE_SIZE* sizes) ;

/** Changes the ordinal counter for the current opponent.
 *
 * @param vm is the virtual machine to use.
 * @param counter is the new value of the ordinal counter.
 */
void VM_setOrdinalCounter(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS counter) ;

/** Replies the ordinal counter for the current opponent.
 *
 * @param vm is the virtual machine to use.
 * @return an ordinal counter
 */
MEMORY_ADDRESS VM_getOrdinalCounter(VIRTUAL_MACHINE* vm) ;

/** Replies the current playing opponent.
 *
 * @param vm is the virtual machine to use.
 * @return the current playing opponent
 */
OPPONENT_SIZE VM_getCurrentPlayer(VIRTUAL_MACHINE* vm) ;

/** Replies the ordinal counter for the specified opponent.
 *
 * @param vm is the virtual machine to use.
 * @param nopponent is the index of the opponent
 * @return an ordinal counter
 */
MEMORY_ADDRESS VM_getOrdinalCounterFor(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent) ;

/** Replies the result of the previously saved test.
 *
 * @param vm is the virtual machine to use.
 * @param result is the result
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getLastTest(VIRTUAL_MACHINE* vm, boolean* result) ;

/** Set the result of the last test.
 *
 * @param vm is the virtual machine to use.
 * @param result is the result
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_setLastTest(VIRTUAL_MACHINE* vm, boolean result) ;

/** Replies the base address of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param base is the base address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getArrayBaseAddress(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* base) ;

/** Replies the addresses of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param baseadr is the base address
 * @param endadr is the end address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getArrayAddresses(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* baseadr, MEMORY_ADDRESS* endadr) ;

/** Replies the addresses of the code segment.
 *
 * @param vm is the virtual machine to use.
 * @param baseadr is the base address
 * @param endadr is the end address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getCodeSegmentAddresses(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* baseadr, MEMORY_ADDRESS* endadr) ;

/** Sets the base address of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param length is the length of the array
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_computeArrayBaseAddress(VIRTUAL_MACHINE* vm, MEMORY_SIZE length) ;

/** Replies the address to the virtual memory.
 *
 * @param vm is the virtual machine to use
 * @return the address of the current virtual memory
 */
VIRTUAL_MEMORY* VM_getVirtualMemory(VIRTUAL_MACHINE* vm) ;

/** Replies the winner of the last play.
 *
 * @param vm is the virtual machine to use.
 * @return the index of the winner.
 */
OPPONENT_SIZE VM_getWinner(VIRTUAL_MACHINE* vm) ;

/** Run the virtual machine.
 *
 * @param vm is the virtual machine to use.
 * @return TRUE if an opponent win, otherwhise FALSE
 */
boolean VM_run(VIRTUAL_MACHINE* vm) ;

/** Sets the error message for the current opponent.
 * The message will NOT be freed by the destructor of
 * the virtual machine.
 *
 * @param vm is the virtual machine to delete.
 * @param message is the error message.
 */
void VM_setErrorMsg(VIRTUAL_MACHINE* vm, ERRORNUM message) ;

/** Replies the error number for the specified opponent.
 *
 * @param vm is the virtual machine to delete.
 * @param n is the index of the opponent
 * @return the error number.
 */
ERRORNUM VM_getError(VIRTUAL_MACHINE* vm, OPPONENT_SIZE opponent) ;

/** Replies the error message for the specified opponent.
 *
 * @param vm is the virtual machine to delete.
 * @param n is the index of the opponent
 * @return the error message.
 */
const char* VM_getErrorMsg(VIRTUAL_MACHINE* vm, OPPONENT_SIZE opponent) ;

/** Stop the current opponent.
 *
 * @param vm is the virtual machine to delete.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_stopOpponent(VIRTUAL_MACHINE* vm) ;

/** Replies the quantity of turns in the last game.
 *
 * @return the quantity of turns.
 */
unsigned long long VM_getTurns(VIRTUAL_MACHINE* vm) ;

#endif
